from datetime import timedelta
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union, cast
from uuid import UUID, uuid4

import pendulum

from prefect._internal.pydantic import HAS_PYDANTIC_V2

if HAS_PYDANTIC_V2:
    from pydantic.v1 import Extra, Field, PrivateAttr, root_validator, validator
    from pydantic.v1.fields import ModelField
else:
    from pydantic import Extra, Field, PrivateAttr, root_validator, validator
    from pydantic.fields import ModelField

from prefect._internal.schemas.bases import PrefectBaseModel
from prefect._internal.schemas.fields import DateTimeTZ
from prefect._internal.schemas.transformations import FieldFrom, copy_model_fields
from prefect.events.actions import ActionTypes, RunDeployment
from prefect.utilities.collections import AutoEnum

# These are defined by Prefect Cloud
MAXIMUM_LABELS_PER_RESOURCE = 500
MAXIMUM_RELATED_RESOURCES = 500


class Posture(AutoEnum):
    Reactive = "Reactive"
    Proactive = "Proactive"


class ResourceSpecification(PrefectBaseModel):
    __root__: Dict[str, Union[str, List[str]]]


class Labelled(PrefectBaseModel):
    """An object defined by string labels and values"""

    __root__: Dict[str, str]

    def keys(self) -> Iterable[str]:
        return self.__root__.keys()

    def items(self) -> Iterable[Tuple[str, str]]:
        return self.__root__.items()

    def __getitem__(self, label: str) -> str:
        return self.__root__[label]

    def __setitem__(self, label: str, value: str) -> str:
        self.__root__[label] = value
        return value


class Resource(Labelled):
    """An observable business object of interest to the user"""

    @root_validator(pre=True)
    def enforce_maximum_labels(cls, values: Dict[str, Any]):
        labels = values.get("__root__")
        if not isinstance(labels, dict):
            return values

        if len(labels) > MAXIMUM_LABELS_PER_RESOURCE:
            raise ValueError(
                "The maximum number of labels per resource "
                f"is {MAXIMUM_LABELS_PER_RESOURCE}"
            )

        return values

    @root_validator(pre=True)
    def requires_resource_id(cls, values: Dict[str, Any]):
        labels = values.get("__root__")
        if not isinstance(labels, dict):
            return values

        labels = cast(Dict[str, str], labels)

        if "prefect.resource.id" not in labels:
            raise ValueError("Resources must include the prefect.resource.id label")
        if not labels["prefect.resource.id"]:
            raise ValueError("The prefect.resource.id label must be non-empty")

        return values

    @property
    def id(self) -> str:
        return self["prefect.resource.id"]


class RelatedResource(Resource):
    """A Resource with a specific role in an Event"""

    @root_validator(pre=True)
    def requires_resource_role(cls, values: Dict[str, Any]):
        labels = values.get("__root__")
        if not isinstance(labels, dict):
            return values

        labels = cast(Dict[str, str], labels)

        if "prefect.resource.role" not in labels:
            raise ValueError(
                "Related Resources must include the prefect.resource.role label"
            )
        if not labels["prefect.resource.role"]:
            raise ValueError("The prefect.resource.role label must be non-empty")

        return values

    @property
    def role(self) -> str:
        return self["prefect.resource.role"]


class Event(PrefectBaseModel):
    """The client-side view of an event that has happened to a Resource"""

    occurred: DateTimeTZ = Field(
        default_factory=pendulum.now,
        description="When the event happened from the sender's perspective",
    )
    event: str = Field(
        description="The name of the event that happened",
    )
    resource: Resource = Field(
        description="The primary Resource this event concerns",
    )
    related: List[RelatedResource] = Field(
        default_factory=list,
        description="A list of additional Resources involved in this event",
    )
    payload: Dict[str, Any] = Field(
        default_factory=dict,
        description="An open-ended set of data describing what happened",
    )
    id: UUID = Field(
        default_factory=uuid4,
        description="The client-provided identifier of this event",
    )
    follows: Optional[UUID] = Field(
        None,
        description=(
            "The ID of an event that is known to have occurred prior to this "
            "one. If set, this may be used to establish a more precise "
            "ordering of causally-related events when they occur close enough "
            "together in time that the system may receive them out-of-order."
        ),
    )

    @property
    def involved_resources(self) -> Iterable[Resource]:
        return [self.resource] + list(self.related)

    @validator("related")
    def enforce_maximum_related_resources(cls, value: List[RelatedResource]):
        if len(value) > MAXIMUM_RELATED_RESOURCES:
            raise ValueError(
                "The maximum number of related resources "
                f"is {MAXIMUM_RELATED_RESOURCES}"
            )

        return value


class Trigger(PrefectBaseModel):
    """Defines the criteria for the events and conditions under which an Automation
    will trigger an action"""

    match: ResourceSpecification = Field(
        default_factory=lambda: ResourceSpecification(__root__={}),
        description="Labels for resources which this Automation will match.",
    )
    match_related: ResourceSpecification = Field(
        default_factory=lambda: ResourceSpecification(__root__={}),
        description="Labels for related resources which this Automation will match.",
    )

    after: Set[str] = Field(
        default_factory=set,
        description=(
            "The event(s) which must first been seen to start this automation.  If "
            "empty, then start this Automation immediately.  Events may include "
            "trailing wildcards, like `prefect.flow-run.*`"
        ),
    )
    expect: Set[str] = Field(
        default_factory=set,
        description=(
            "The event(s) this automation is expecting to see.  If empty, this "
            "automation will match any event.  Events may include trailing wildcards, "
            "like `prefect.flow-run.*`"
        ),
    )

    for_each: Set[str] = Field(
        default_factory=set,
        description=(
            "Evaluate the Automation separately for each distinct value of these labels"
            " on the resource"
        ),
    )
    posture: Posture = Field(
        Posture.Reactive,
        description=(
            "The posture of this Automation, either Reactive or Proactive.  Reactive "
            "automations respond to the _presence_ of the expected events, while "
            "Proactive  automations respond to the _absence_ of those expected events."
        ),
    )
    threshold: int = Field(
        1,
        description=(
            "The number of events required for this Automation to trigger (for "
            "Reactive automations), or the number of events expected (for Proactive "
            "automations)"
        ),
    )
    within: timedelta = Field(
        timedelta(0),
        minimum=0.0,
        exclusiveMinimum=False,
        description=(
            "The time period over which the events must occur.  For Reactive triggers, "
            "this may be as low as 0 seconds, but must be at least 10 seconds for "
            "Proactive triggers"
        ),
    )

    @validator("within")
    def enforce_minimum_within(
        cls, value: timedelta, values, config, field: ModelField
    ):
        minimum = field.field_info.extra["minimum"]
        if value.total_seconds() < minimum:
            raise ValueError("The minimum within is 0 seconds")
        return value

    @root_validator(skip_on_failure=True)
    def enforce_minimum_within_for_proactive_triggers(cls, values: Dict[str, Any]):
        posture: Optional[Posture] = values.get("posture")
        within: Optional[timedelta] = values.get("within")

        if posture == Posture.Proactive:
            if not within or within == timedelta(0):
                values["within"] = timedelta(seconds=10.0)
            elif within < timedelta(seconds=10.0):
                raise ValueError(
                    "The minimum within for Proactive triggers is 10 seconds"
                )

        return values


class Automation(PrefectBaseModel):
    """Defines an action a user wants to take when a certain number of events
    do or don't happen to the matching resources"""

    class Config:
        extra = Extra.ignore

    name: str = Field(..., description="The name of this automation")
    description: str = Field("", description="A longer description of this automation")

    enabled: bool = Field(True, description="Whether this automation will be evaluated")

    trigger: Trigger = Field(
        ...,
        description=(
            "The criteria for which events this Automation covers and how it will "
            "respond to the presence or absence of those events"
        ),
    )

    actions: List[ActionTypes] = Field(
        ...,
        description="The actions to perform when this Automation triggers",
    )
    owner_resource: Optional[str] = Field(
        default=None, description="The owning resource of this automation"
    )


class ExistingAutomation(Automation):
    id: UUID = Field(..., description="The ID of this automation")


@copy_model_fields
class ResourceTrigger(PrefectBaseModel):
    name: Optional[str] = Field(
        None, description="The name to give to the automation created for this trigger."
    )
    description: str = FieldFrom(Automation)
    enabled: bool = FieldFrom(Automation)

    match: ResourceSpecification = FieldFrom(Trigger)
    match_related: ResourceSpecification = FieldFrom(Trigger)
    after: Set[str] = FieldFrom(Trigger)
    expect: Set[str] = FieldFrom(Trigger)
    for_each: Set[str] = FieldFrom(Trigger)
    posture: Posture = FieldFrom(Trigger)
    threshold: int = FieldFrom(Trigger)
    within: timedelta = FieldFrom(Trigger)

    def as_automation(self) -> Automation:
        assert self.name
        return Automation(
            name=self.name,
            description=self.description,
            enabled=self.enabled,
            trigger=Trigger(
                match=self.match,
                match_related=self.match_related,
                after=self.after,
                expect=self.expect,
                for_each=self.for_each,
                posture=self.posture,
                threshold=self.threshold,
                within=self.within,
            ),
            actions=self.actions(),
            owner_resource=self.owner_resource(),
        )

    def owner_resource(self) -> Optional[str]:
        return None

    def actions(self) -> List[ActionTypes]:
        raise NotImplementedError


@copy_model_fields
class DeploymentTrigger(ResourceTrigger):
    _deployment_id: Optional[UUID] = PrivateAttr(default=None)
    parameters: Optional[Dict[str, Any]] = FieldFrom(RunDeployment)

    def set_deployment_id(self, deployment_id: UUID):
        self._deployment_id = deployment_id

    def owner_resource(self) -> Optional[str]:
        return f"prefect.deployment.{self._deployment_id}"

    def actions(self) -> List[ActionTypes]:
        assert self._deployment_id
        return [
            RunDeployment(
                parameters=self.parameters,
                deployment_id=self._deployment_id,
            )
        ]



"""
Core primitives for running Prefect project steps.

Project steps are YAML representations of Python functions along with their inputs.

Whenever a step is run, the following actions are taken:

- The step's inputs and block / variable references are resolved (see [the projects concepts documentation](/concepts/projects/#templating-options) for more details)
- The step's function is imported; if it cannot be found, the `requires` keyword is used to install the necessary packages
- The step's function is called with the resolved inputs
- The step's output is returned and used to resolve inputs for subsequent steps
"""
import os
import re
import subprocess
import warnings
from copy import deepcopy
from importlib import import_module
from typing import Any, Dict, List, Optional, Tuple

from prefect._internal.compatibility.deprecated import PrefectDeprecationWarning
from prefect._internal.concurrency.api import Call, from_async
from prefect.logging.loggers import get_logger
from prefect.settings import PREFECT_DEBUG_MODE
from prefect.utilities.importtools import import_object
from prefect.utilities.processutils import get_sys_executable
from prefect.utilities.templating import (
    apply_values,
    resolve_block_document_references,
    resolve_variables,
)

RESERVED_KEYWORDS = {"id", "requires"}


class StepExecutionError(Exception):
    """
    Raised when a step fails to execute.
    """


def _strip_version(requirement: str) -> str:
    """
    Strips the version from a requirement string.

    Args:
        requirement: A requirement string, e.g. "requests>=2.0.0"

    Returns:
        The package name, e.g. "requests"

    Examples:
        ```python
        >>> _strip_version("s3fs>=2.0.0<3.0.0")
        "s3fs"
        ```
    """
    # split on any of the characters in the set [<>=!~]
    # and return the first element which will be the package name
    return re.split(r"[<>=!~]", requirement)[0].strip()


def _get_function_for_step(fully_qualified_name: str, requires: Optional[str] = None):
    if not isinstance(requires, list):
        packages = [requires] if requires else []
    else:
        packages = requires

    try:
        for package in packages:
            import_module(_strip_version(package))
        step_func = import_object(fully_qualified_name)
        return step_func
    except ImportError:
        if requires:
            print(
                f"Unable to load step function: {fully_qualified_name}. Attempting"
                f" install of {requires}."
            )
        else:
            raise

    subprocess.check_call(
        [get_sys_executable(), "-m", "pip", "install", ",".join(packages)]
    )
    step_func = import_object(fully_qualified_name)
    return step_func


async def run_step(step: Dict, upstream_outputs: Optional[Dict] = None) -> Dict:
    """
    Runs a step, returns the step's output.

    Steps are assumed to be in the format `{"importable.func.name": {"kwarg1": "value1", ...}}`.

    The 'id and 'requires' keywords are reserved for specific purposes and will be removed from the
    inputs before passing to the step function:

    This keyword is used to specify packages that should be installed before running the step.
    """
    fqn, inputs = _get_step_fully_qualified_name_and_inputs(step)
    upstream_outputs = upstream_outputs or {}

    if len(step.keys()) > 1:
        raise ValueError(
            f"Step has unexpected additional keys: {', '.join(list(step.keys())[1:])}"
        )

    keywords = {
        keyword: inputs.pop(keyword)
        for keyword in RESERVED_KEYWORDS
        if keyword in inputs
    }

    inputs = apply_values(inputs, upstream_outputs)
    inputs = await resolve_block_document_references(inputs)
    inputs = await resolve_variables(inputs)
    inputs = apply_values(inputs, os.environ)
    step_func = _get_function_for_step(fqn, requires=keywords.get("requires"))
    result = await from_async.call_soon_in_new_thread(
        Call.new(step_func, **inputs)
    ).aresult()
    return result


async def run_steps(
    steps: List[Dict[str, Any]],
    upstream_outputs: Optional[Dict[str, Any]] = None,
    print_function: Any = print,
):
    upstream_outputs = deepcopy(upstream_outputs) if upstream_outputs else {}
    for step in steps:
        if not step:
            continue
        fqn, inputs = _get_step_fully_qualified_name_and_inputs(step)
        step_name = fqn.split(".")[-1]
        print_function(f" > Running {step_name} step...")
        try:
            # catch warnings to ensure deprecation warnings are printed
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter(
                    "always",
                    category=PrefectDeprecationWarning,
                )
                warnings.simplefilter(
                    "always",
                    category=DeprecationWarning,
                )
                step_output = await run_step(step, upstream_outputs)
            if w:
                printed_messages = []
                for warning in w:
                    message = str(warning.message)
                    # prevent duplicate warnings from being printed
                    if message not in printed_messages:
                        try:
                            # try using rich styling
                            print_function(message, style="yellow")
                        except Exception:
                            # default to printing without styling
                            print_function(message)
                        printed_messages.append(message)

            if not isinstance(step_output, dict):
                if PREFECT_DEBUG_MODE:
                    get_logger().warning(
                        "Step function %s returned unexpected type: %s",
                        fqn,
                        type(step_output),
                    )
                continue
            # store step output under step id to prevent clobbering
            if inputs.get("id"):
                upstream_outputs[inputs.get("id")] = step_output
            upstream_outputs.update(step_output)
        except Exception as exc:
            raise StepExecutionError(f"Encountered error while running {fqn}") from exc
    return upstream_outputs


def _get_step_fully_qualified_name_and_inputs(step: Dict) -> Tuple[str, Dict]:
    step = deepcopy(step)
    return step.popitem()



# Note: This file should only be imported under a HAS_PYDANTIC_V2 flag

import inspect
import typing as t

import pendulum
import pydantic
from pydantic import BaseModel as V2BaseModel
from pydantic import ConfigDict, create_model
from pydantic.type_adapter import TypeAdapter

from prefect._internal.pydantic.annotations.pendulum import (
    PydanticPendulumDateTimeType,
    PydanticPendulumDateType,
    PydanticPendulumDurationType,
)
from prefect._internal.pydantic.schemas import GenerateEmptySchemaForUserClasses


def has_v2_model_as_param(signature: inspect.Signature) -> bool:
    parameters = signature.parameters.values()
    return any(
        isinstance(p, V2BaseModel)
        or (inspect.isclass(p.annotation) and issubclass(p.annotation, V2BaseModel))
        for p in parameters
    )


def process_v2_params(
    param: inspect.Parameter,
    *,
    position: int,
    docstrings: t.Dict[str, str],
    aliases: t.Dict
) -> t.Tuple[str, t.Any, "pydantic.Field"]:
    """
    Generate a sanitized name, type, and pydantic.Field for a given parameter.

    This implementation is exactly the same as the v1 implementation except
    that it uses pydantic v2 constructs.
    """
    # Pydantic model creation will fail if names collide with the BaseModel type
    if hasattr(pydantic.BaseModel, param.name):
        name = param.name + "__"
        aliases[name] = param.name
    else:
        name = param.name

    type_ = t.Any if param.annotation is inspect._empty else param.annotation

    # Replace pendulum type annotations with our own so that they are pydantic compatible
    if type_ == pendulum.DateTime:
        type_ = PydanticPendulumDateTimeType
    if type_ == pendulum.Date:
        type_ = PydanticPendulumDateType
    if type_ == pendulum.Duration:
        type_ = PydanticPendulumDurationType

    field = pydantic.Field(
        default=... if param.default is param.empty else param.default,
        title=param.name,
        description=docstrings.get(param.name, None),
        alias=aliases.get(name),
        json_schema_extra={"position": position},
    )
    return name, type_, field


def create_v2_schema(name_: str, model_cfg: ConfigDict, **model_fields):
    """
    Create a pydantic v2 model and craft a v1 compatible schema from it.
    """
    model = create_model(name_, __config__=model_cfg, **model_fields)
    adapter = TypeAdapter(model)

    # root model references under #definitions
    schema = adapter.json_schema(
        by_alias=True,
        ref_template="#/definitions/{model}",
        schema_generator=GenerateEmptySchemaForUserClasses,
    )
    # ensure backwards compatability by copying $defs into definitions
    if "$defs" in schema:
        schema["definitions"] = schema["$defs"]
    return schema



import pendulum
import uvicorn
from prefect._vendor.fastapi import APIRouter, FastAPI, status
from prefect._vendor.fastapi.responses import JSONResponse

from prefect.settings import (
    PREFECT_RUNNER_POLL_FREQUENCY,
    PREFECT_RUNNER_SERVER_HOST,
    PREFECT_RUNNER_SERVER_LOG_LEVEL,
    PREFECT_RUNNER_SERVER_MISSED_POLLS_TOLERANCE,
    PREFECT_RUNNER_SERVER_PORT,
)


def perform_health_check(runner, delay_threshold: int = None) -> JSONResponse:
    if delay_threshold is None:
        delay_threshold = (
            PREFECT_RUNNER_SERVER_MISSED_POLLS_TOLERANCE.value()
            * PREFECT_RUNNER_POLL_FREQUENCY.value()
        )

    def _health_check():
        now = pendulum.now("utc")
        poll_delay = (now - runner.last_polled).total_seconds()

        if poll_delay > delay_threshold:
            return JSONResponse(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                content={"message": "Runner is unresponsive at this time"},
            )
        return JSONResponse(status_code=status.HTTP_200_OK, content={"message": "OK"})

    return _health_check


def run_count(runner) -> int:
    def _run_count():
        run_count = len(runner._flow_run_process_map)
        return run_count

    return _run_count


def shutdown(runner) -> int:
    def _shutdown():
        runner.stop()
        return JSONResponse(status_code=status.HTTP_200_OK, content={"message": "OK"})

    return _shutdown


def start_webserver(
    runner,
    log_level: str = None,
) -> None:
    """
    Run a FastAPI server for a runner.

    Args:
        runner (Runner): the runner this server interacts with and monitors
        log_level (str): the log level to use for the server
    """
    webserver = FastAPI()
    router = APIRouter()

    router.add_api_route(
        "/health", perform_health_check(runner=runner), methods=["GET"]
    )
    router.add_api_route("/run_count", run_count(runner=runner), methods=["GET"])
    router.add_api_route("/shutdown", shutdown(runner=runner), methods=["POST"])

    webserver.include_router(router)

    host = PREFECT_RUNNER_SERVER_HOST.value()
    port = PREFECT_RUNNER_SERVER_PORT.value()
    log_level = log_level or PREFECT_RUNNER_SERVER_LOG_LEVEL.value()

    uvicorn.run(webserver, host=host, port=port, log_level=log_level)
